/*
 *      Author: (C) Patrick Schäfer
 *      		patrick.schaefer@zib.de
 */
#include "classify.h"

#include <time.h>
#include <stdio.h>
#ifdef _OPENMP
# include <omp.h>
#endif
#include <unordered_map>
#include <map>
#include <queue>
#include <set>
#include <iomanip>
#include <limits>
#include <iostream>
#include <iterator>
#include <algorithm>
#include <string>
#include <sstream>
#include <array>
#include <utility>
#include <limits.h>
#include <sparsehash/dense_hash_map>

#define abs(x)   ((x) >0 ?(x):-(x))

#define MAX_CHAR_PER_LINE 200000

#define DEBUG 0
#define DOUBLE_MAX_VALUE std::numeric_limits<double>::infinity()

const int maxF = 16;
const int minF = 4;
const int maxS = 4;
short usedBits = 2;

const int splits = 10;
const uint MAX_WINDOW_LENGTH = 4000;

// the best window-sizes to choose from
const double factor = 1.0;

std::vector< std::shared_ptr<TimeSeries> > trainSamples;
std::vector< std::shared_ptr<TimeSeries> > testSamples;

double testTime = 0.0;
double trainTime = 0.0;

std::string datasets[] = {
  "Two_Patterns",
  "ChlorineConcentration",
  "wafer",
  "MedicalImages",
  "FaceAll",
  "OSULeaf",
  "Adiac",
  "SwedishLeaf",
  "yoga",
  "fish",
  "Lighting7",
  "Lighting2",
  "Trace",
  "synthetic_control",
  "FacesUCR",
  "CinC_ECG_torso",
  "MALLAT",
  "Symbols",
  "Coffee",
  "ECG200",
  "FaceFour",
  "OliveOil",
  "Gun_Point",
  "Beef",
  "DiatomSizeReduction",
  "CBF",
  "ECGFiveDays",
  "TwoLeadECG",
  "SonyAIBORobot SurfaceII",
  "MoteStrain",
  "ItalyPowerDemand",
  "SonyAIBORobot Surface",
  "Haptics",
  "InlineSkate",
  "50words",
  "Cricket_Y",
  "Cricket_X",
  "Cricket_Z",
  "WordsSynonyms",
  "uWaveGestureLibrary_Z",
  "uWaveGestureLibrary_Y",
  "uWaveGestureLibrary_X",
  "NonInvasiveFatalECG_Thorax1",
  "NonInvasiveFatalECG_Thorax2",
  "StarLightCurves",

};

void loadDatset(const char* filename, std::vector< std::shared_ptr<TimeSeries> > & samples) {
  
  FILE* f;
  f = fopen(filename, "r");
  if (f==NULL) {
	std::cout << "Error!! Input file " << filename << " not found." << std::endl;
	exit(0);
  }
  
  char buf[MAX_CHAR_PER_LINE];
  char* tmp;
  double classLabel = 0;
  
  while (fgets(buf, MAX_CHAR_PER_LINE, f)) {
	std::vector<double> tsData;
	
	// split buf into tokens
	tmp = strtok(buf, " \r\n");
	
	// read label
	classLabel = atof(tmp);
	
	// read data
	tmp = strtok(NULL, " \r\n");
	while(tmp != NULL) {
	  tsData.push_back(atof(tmp));
	  tmp = strtok(NULL," \r\n");
	}
	
	std::shared_ptr<TimeSeries> ts = std::make_shared<TimeSeries>(tsData, classLabel);
	ts->norm(true);
	samples.push_back(ts);
	
  }
  
  std::cout << "Done reading: " << filename
  << ". " << samples.size()
  << " time series of length "
  << samples[0]->getSize() << std::endl;
  
  fclose(f);
}

std::string classify() {
  std::chrono::time_point<std::chrono::system_clock> startTime = std::chrono::system_clock::now();
  
  int correct = predictEuclideanDistance(testSamples, trainSamples);
  outputResult(correct, startTime, testSamples.size());
  
  // training
  int minWindowLength = 10;
  int maxWindowLength = std::min(MAX_WINDOW_LENGTH,trainSamples[0]->getSize());
  
  int bestWindowLength = 0;
  bool bestNormedMean = false;
  int bestCorrectTraining = 0;
  std::vector<WindowSizes> bestWindows;
  
  // training
  for (auto normWindows : {true, false}) {
	startTime = std::chrono::system_clock::now();
	
	int currentCorrect = 0;
	int currentWindowLength = 0;
	std::vector<WindowSizes> windows = fit(maxWindowLength, minWindowLength, currentCorrect, currentWindowLength, normWindows);
	
	std::cout << std::endl << "Training BOSS VS:\t"
	  << "best window-size:\t" << currentWindowLength
	  << "\t normed subsequences: \t"
	  << (normWindows? "true":"false") << std::endl;
	outputResult(currentCorrect, startTime, trainSamples.size());
	trainTime += std::chrono::duration_cast<std::chrono::milliseconds> (std::chrono::system_clock::now() - startTime).count() / 1000.0;
	
	if (bestCorrectTraining <= currentCorrect) {
	  bestCorrectTraining = currentCorrect;
	  bestNormedMean = normWindows;
	  bestWindowLength = currentWindowLength;
	  bestWindows = windows;
	}
  }

  // testing
  std::cout << std::endl << "Testing BOSS VS:\t"
	<< "best window-size:\t" << bestWindowLength
	<< "\t normed subsequences: \t"
	<< (bestNormedMean? "true":"false") << std::endl;

  // boss vs classifier
  startTime = std::chrono::system_clock::now();
  
  int correctTesting = predictEnsemble(bestWindows, bestCorrectTraining, bestNormedMean, testSamples, trainSamples);
  outputResult(correctTesting, startTime, testSamples.size());
  testTime += std::chrono::duration_cast<std::chrono::milliseconds> (std::chrono::system_clock::now() - startTime).count() / 1000.0;
  
  double testingCorrect = 1-formatError(correctTesting, testSamples.size());
  double trainingCorrect = 1-formatError(bestCorrectTraining, trainSamples.size());
  
  std::stringstream msg;
  msg << trainingCorrect << ";"
  << testingCorrect << ";"
  << (bestNormedMean? "true":"false") << ";"
  << bestWindowLength;
  
  return msg.str();
}

void printArray(uint* data, size_t size) {
  std::copy(data, data + size, std::ostream_iterator<int>(std::cout, " "));
  std::cout << std::endl;
}

void printArray(int* data, size_t size) {
  std::copy(data, data + size, std::ostream_iterator<int>(std::cout, " "));
  std::cout << std::endl;
}

void printArray(double* data, int size) {
  std::copy(data, data + size, std::ostream_iterator<double>(std::cout, " "));
  std::cout << std::endl;
}

int predictEuclideanDistance(const std::vector< std::shared_ptr<TimeSeries> > & testSamples,
							 const std::vector< std::shared_ptr<TimeSeries> > & trainSamples) {
  int correct = 0;
  
  for (auto query : testSamples) {
	double bestSoFar = DOUBLE_MAX_VALUE;
	double label = -1;
	for (auto ts : trainSamples) {
	  if (ts != query) {
		double distance = getEuclideanDistance(ts, query, bestSoFar);
		if (distance < bestSoFar) {
		  bestSoFar = distance;
		  label = ts->getLabel();
		}
	  }
	}
	if (query->getLabel() == label) {
	  correct++;
	}
  }
  return correct;
}

double getEuclideanDistance(std::shared_ptr<TimeSeries> ts,
							std::shared_ptr<TimeSeries> q,
							double minValue) {
  
  double distance = 0.0;
  double* tsData = ts->getData();
  double* qData = q->getData();
  for (uint ww = 0; ww < ts->getSize(); ww++) {
	double value = qData[ww] - tsData[ww];
	distance += value*value;
	
	// early abandoning
	if (distance > minValue) {
	  return DOUBLE_MAX_VALUE;
	}
  }
  return distance;
}

std::vector<WindowSizes> fit(int maxWindowLength,
							 int minWindowLength,
							 int& bestAcc,
							 int& bestWindowSize,
							 bool norm) {
  
  // build sets for leave-one-out cross validation
  std::vector< std::vector<int> > sets = getStratifiedTrainTestSplitIndices(trainSamples);
  
  std::vector< std::vector<int> > trainIndices(splits);
  std::vector< std::vector<int> > testIndices(splits);
  
  for (int s = 0; s < splits; s++) {
	testIndices[s] = sets[s];
	trainIndices[s] = getSplit(sets, s);
  }
  
  std::vector<WindowSizes> windows;
  volatile int bestAccuracy = 0;
  
  const std::set<double> uniqueLabels = uniqueClassLabels(trainSamples);
  
#pragma omp parallel
  {
	std::vector< std::vector<int> > wordsTrain (trainSamples.size());
	std::vector<BagOfPattern> bagTrain(trainSamples.size());
	
	google::dense_hash_map< double, std::shared_ptr<Statistic> > matrix;
	matrix.set_empty_key(DOUBLE_MAX_VALUE);
	
	// subset of windows
	double count = sqrt(maxWindowLength);
	int distance = (int)((maxWindowLength-minWindowLength)/count);

	#pragma omp for schedule(dynamic)
	for (int i = 0; i <= (int)(count); i++) {
	  int windowLength = minWindowLength+i*distance;
	  
	  // all windows
	  // for (int windowLength = maxWindowLength; windowLength >= minWindowLength; windowLength--) {
	  
	  int bestCorrect = 0;
	  int bestFeatures = 0;
	  int bestSymbols = 0;
	  
	  std::shared_ptr<SFA> signature = createWords(windowLength,
								   std::min(windowLength+windowLength%2,maxF),
								   maxS,
								   trainSamples,
								   wordsTrain,
								   NULL,
								   norm);
	  
	  // use several test splits
	  int length = (maxF - minF) / 2 + 1;
	  int scores[length];
	  std::fill_n( scores, length, 0 );
	  
	  for (int s = 0; s < splits; s++) {
		int pos = 0;
		
		for (int f = minF; f <= std::min(windowLength,maxF); f+=2) {
		  createBagOfPatterns(bagTrain,
							  trainSamples,
							  wordsTrain,
							  f,
							  maxS);
		  
		  calcTfIdf(bagTrain,
					&matrix,
//					wordInClassFreq,
					trainSamples,
					trainIndices[s],
					uniqueLabels,
					wordsTrain,
					f,
					maxS);
		  
		  normalizeStatistics(matrix);
		  
		  Predictions p = predict(trainSamples,
								  testIndices[s],
								  bagTrain,
								  matrix);
		  
		  scores[pos++] += p.correct;
		}
		
	  }
	  
	  
	  for (int i = 0; i < length; i++) {
		if (scores[i] > bestCorrect) {
		  bestCorrect = scores[i];
		  bestFeatures = minF+i*2;
		  bestSymbols = maxS;
		}
	  }
	  
	  
	  if (bestCorrect > bestAccuracy) {
#pragma omp critical (BESTCORRECT)
		{
		  if (bestCorrect > bestAccuracy) {
			bestAccuracy = bestCorrect;
			bestWindowSize = windowLength;
		  }
		}
	  }
	  
	  if (bestCorrect >= bestAccuracy*factor) {
		windows.push_back(WindowSizes(bestCorrect,
									  windowLength,
									  bestSymbols,
									  bestFeatures,
									  norm,
									  signature,
									  wordsTrain));
	  }
	}
  }
  
  
  for (auto best : windows) {
	if (best.accuracy < bestAccuracy*factor) {
	  best.clear();
	}
  }
  
  bestAcc = bestAccuracy;
  
  return windows;
}

int predictEnsemble(const std::vector<WindowSizes>& results,
					int bestKnownCorrect,
					bool normMean,
					const std::vector< std::shared_ptr<TimeSeries> > & testSamples,
					const std::vector< std::shared_ptr<TimeSeries> > & trainSamples) {
  
  std::vector<std::vector<double> > testLabels;
  testLabels.resize(testSamples.size());
  
  std::vector<uint> currentWindowLengths;
  currentWindowLengths.reserve(results.size());
  
  const std::set<double> uniqueLabels = uniqueClassLabels(trainSamples);
  
  std::vector<int> indicesTest(testSamples.size());
  for (unsigned int i = 0; i < testSamples.size(); i++) {
	indicesTest[i]=i;
  }
  
  std::vector<int> indicesTrain(trainSamples.size());
  for (unsigned int i = 0; i < trainSamples.size(); i++) {
	indicesTrain[i]=i;
  }
  
  // determine the labels for each window
  // iterate each sample to classify
#pragma omp parallel
  {
	google::dense_hash_map< double, std::shared_ptr<Statistic> > matrix;
	matrix.set_empty_key(DOUBLE_MAX_VALUE);
	
	std::vector<BagOfPattern> bagTrain(trainSamples.size());
	std::vector<BagOfPattern> bagTest(testSamples.size());
	
	std::vector< std::vector<int> > wordsTest (testSamples.size());
	
#pragma omp for
	for (uint i = 0; i < results.size(); ++i) {
	  
	  WindowSizes window = results[i];
	  if ((double)window.accuracy >= ((double)bestKnownCorrect * factor)) { // all with a better score
		
		currentWindowLengths.push_back(window.windowSize);
		
		// create sliding windows in parallel
		createWords(window.windowSize,
					window.features,
					window.symbols,
					testSamples,
					wordsTest,
					window.signature,
					window.normed);
		
		createBagOfPatterns(bagTest,
							testSamples,
							wordsTest,
							window.features,
							window.symbols);
		
		createBagOfPatterns(bagTrain,
							trainSamples,
							window.words,
							window.features,
							window.symbols);
		
		calcTfIdf(bagTrain,
				  &matrix,
				  trainSamples,
				  indicesTrain,
				  uniqueLabels,
				  window.words,
				  window.features,
				  window.symbols);
		
		normalizeStatistics(matrix);
		
		Predictions p = predict(testSamples,
								indicesTest,
								bagTest,
								matrix);
		
#pragma omp critical (LABELS)
		{
		  for (uint i = 0; i < testSamples.size(); ++i) {
			testLabels[i].emplace_back(p.labels[i]);
		  }
		}
	  }
	}
  }
  
//  std::cout << "There are " << currentWindowLengths.size() << " models" << std::endl;
//  for (auto c : currentWindowLengths) {
//	std::cout << c << std::endl;
//  }
  
  int correct = 0;
  
  // choose the most frequent labels for each sample
  for (uint q = 0; q < testLabels.size(); q++) {
	
	long maxCount = 0l;
	double maxLabel = -1;
	std::unordered_map<double, long> counts;
	
	for (auto label : testLabels[q]) {
	  auto iter = counts.find(label);
	  long count = 0;
	  if (iter == counts.end()) {
		count = 1;
		counts[label] = count;
	  }
	  else {
		count = ++(*iter).second;
	  }
	  
	  if (maxCount < count
		  || (count == maxCount && std::to_string(maxLabel) < std::to_string(label))) {
		maxCount = count;
		maxLabel = label;
	  }
	}
	
	if (maxLabel == testSamples[q]->getLabel()) {
	  correct++;
	}
  }
  
  return correct;
}

const Predictions predict(const std::vector< std::shared_ptr<TimeSeries> > & samples,
						  const std::vector<int>& indices,
						  const std::vector<BagOfPattern>& bagOfPatterns,
						  const google::dense_hash_map< double, std::shared_ptr<Statistic>  >& matrixTrain) {
  
  int correct = 0;
  
  std::vector<double> labels(samples.size());
  
	// iterate each sample to classify
	#pragma omp parallel for schedule(dynamic, 10)
	for (unsigned int a = 0; a < indices.size(); a++) {
	  int i = indices[a];
	  double bestDistance = 0.0;
	  
	  // for each class
	  for (auto& classEntry : matrixTrain) {
		
		double label = classEntry.first;
		auto& tfIdf = classEntry.second->tfIdf;
		
		// determine euclidean distance
		double distance = 0.0;
		for (auto& wordFreq : bagOfPatterns[i].wordFrequencies) {
		  distance += wordFreq.second * (tfIdf[wordFreq.first]+1);
		}
		
		// norm magnitudes??
		
		// update nearest neighbor
		if (distance > bestDistance) {
		  bestDistance = distance;
		  labels[i] = label;
		}
	  }
	  
	  if (samples[i]->getLabel() == labels[i]) {
		__sync_fetch_and_add( &correct, 1 );
	  }
	}
  
  return {labels, correct};
}


std::vector<int> transform(std::shared_ptr<TimeSeries> ts,
							std::shared_ptr<SFA> signature,
							int windowSize,
							int maxFeatures,
							bool normMean) {
  MFT mft(windowSize, normMean, signature);
  return mft.transform(ts, signature->maxCoefficient);
}


std::shared_ptr<SFA> trainHistogram(uint windowSize,
					uint maxFeatures,
					uint maxSymbols,
					const std::vector< std::shared_ptr<TimeSeries> > & trainSamples,
					bool normMean) {
  
  std::shared_ptr<SFA> sfa(new SFA(SFA::EQUI_DEPTH, windowSize, maxFeatures, maxSymbols, normMean));
  
  // create histogramm + distribution
  std::vector<std::shared_ptr<TimeSeries> > windows;
  windows.reserve(trainSamples.size() * trainSamples[0]->getSize() / windowSize);
  
  for (auto t : trainSamples) {
	for (auto t2 : t->getDisjointSequences(windowSize, normMean)) {
	  t2->norm();
	  windows.push_back(t2);
	}
  }
  
  sfa->divideHistogram(windows, 0);
  
  if (DEBUG) {
	printArray(sfa->bestCoefficients, sfa->coefficients);
	sfa->printHistogram();
  }
  
  
  return sfa;
}


std::shared_ptr<SFA> createWords(uint windowSize,
				 uint maxFeatures,
				 uint maxSymbols,
				 const std::vector<std::shared_ptr<TimeSeries>> & samples,
				 std::vector<std::vector <int> > & words,
				 std::shared_ptr<SFA> sig,
				 bool normMean) {
  
  std::shared_ptr<SFA> signature = sig == NULL?
	trainHistogram(windowSize, maxFeatures, maxSymbols, samples, normMean) : sig;
  
  // create sliding windows
  #pragma omp parallel for schedule(dynamic)
  for (uint i = 0; i < samples.size(); i++) {
	words[i] = transform(samples[i], signature, windowSize, maxFeatures, normMean);
	
  }
  return signature;
}

void normalizeStatistics(google::dense_hash_map< double, std::shared_ptr<Statistic> > & classStatistics) {
  
  for (auto & classStat : classStatistics) {
	double squareSum = 0.0;
	for (auto & entry : classStat.second->tfIdf) {
	  squareSum += entry.second*entry.second;
	}
	double squareRoot = std::sqrt(squareSum);
	
	if (squareRoot>0) {
	  for (auto & entry : classStat.second->tfIdf) {
		entry.second /= squareRoot;
	  }
	}
  }
}

const std::set<double> uniqueClassLabels(const std::vector< std::shared_ptr<TimeSeries> > & samples) {
  std::set<double> labels;
  for (auto t : samples) {
	labels.insert(t->getLabel());
  }
  return labels;
}

void initMatrix(google::dense_hash_map< double, std::shared_ptr<Statistic> >* matrix,
				const std::set<double> & uniqueLabels,
				const std::vector< std::vector<int> > & words) {
  if (matrix == NULL || matrix->empty()) {
	for (auto label : uniqueLabels) {
	  (*matrix)[label] = std::make_shared<Statistic>(words.size()*words[0].size());
	}
  }
  else {
	for (auto label : uniqueLabels) {
	  (*matrix)[label]->clear();
	}
  }
}

/**
 * Create all words for a fixed window size
 */
void createBagOfPatterns(std::vector<BagOfPattern>& bagOfPatterns,
						 const std::vector< std::shared_ptr<TimeSeries> > & samples,
						 const std::vector< std::vector <int> > & words,
						 int features,
						 int symbols) {
  
  long mask = (1l << (usedBits*features)) - 1l;
  
  // iterate all samples
#pragma omp parallel for
  for (ushort j = 0; j < samples.size(); j++) {
	bagOfPatterns[j].label = samples[j]->getLabel();
	bagOfPatterns[j].wordFrequencies.clear_no_resize();
	
	// create subsequences
	long lastWord = LONG_MIN;
	for (uint offset = 0; offset < words[j].size(); offset++) {
	  // use the words of larger length to get words of smaller lengths
	  int word = (words[j][offset] & mask);
	  if (word != lastWord && word != INT_MAX) { // ignore adjacent samples
		++bagOfPatterns[j].wordFrequencies[word];
	  }
	  lastWord = word;
	}
  }
}


void calcTfIdf(std::vector<BagOfPattern>& bagOfPatterns,
			   google::dense_hash_map< double, std::shared_ptr<Statistic> >* matrix,
			   const std::vector< std::shared_ptr<TimeSeries> > & samples,
			   const std::vector<int>& indices,
			   const std::set<double> & uniqueLabels,
			   const std::vector<std::vector <int> > & words,
			   int features,
			   int symbols) {
  
  initMatrix(matrix, uniqueLabels, words);
  
  size_t c = 0;
  for (int j : indices) {
	double label = bagOfPatterns[j].label;
	auto& wordInBagFreq = (*matrix)[label]->tfIdf;
	
	for (auto& cursor : bagOfPatterns[j].wordFrequencies) {
	  wordInBagFreq[cursor.first] += cursor.second;
	}
	c = std::max(wordInBagFreq.size(),c);
  }
  
  // count the number of classes where the word is present
  google::dense_hash_map< int, ushort > wordInClassFreq(c);
  wordInClassFreq.set_empty_key(INT_MAX);
  
  for (auto& stat : *matrix) {
	// count the occurence of words
	for (auto& key : stat.second->tfIdf) {
	  ++wordInClassFreq[key.first];
	}
  }
  
  size_t labels = uniqueLabels.size();
  
  // calculate the tfIDF value for each class
  for (auto& stat : *matrix) {
	auto& tfIDFs = stat.second->tfIdf;
	
	// calculate the tfIDF value for each word
	for (auto& patternFrequency : tfIDFs) {
	  auto word = wordInClassFreq[patternFrequency.first];
	  
	  if (word > 0 && labels != word  // avoid log(1)
			) {
		double tfValue = 1.0 + std::log10((double)patternFrequency.second);
		double idfValue = std::log10(1.0+(double)uniqueLabels.size() / (double)word);
		double tfIdf = tfValue / idfValue;
		  
		// update the tfIDF vector
		patternFrequency.second = (float)tfIdf;
	  }
	  else {
		patternFrequency.second = 0;
	  }
	}
  }
}

std::vector< std::vector<int> > getStratifiedTrainTestSplitIndices(const std::vector< std::shared_ptr<TimeSeries> > & samples) {
  
  std::map<double, std::queue<int> > elements;
  
  for (uint i = 0; i < samples.size(); i++) {
	double label = samples[i]->getLabel();
	elements[label].push(i);
  }
  
  double trainTestSplit = 1.0 / splits;
  
  // pick samples
  std::vector< std::vector<int>> sets(splits, std::vector<int>());
  
  // all but one
  for (int s = 0; s < splits-1; s++) {
	for (auto& data : elements) {
	  int count = (int)(data.second.size() * trainTestSplit);
	  int i = 0;
	  while (!data.second.empty() && i <= count) {
		sets[s].push_back(data.second.front());
		data.second.pop();
		i++;
	  }
	}
  }
  
  // add the last split
  for (auto& data : elements) {
	while (!data.second.empty()) {
	  sets[splits-1].push_back(data.second.front());
	  data.second.pop();
	}
  }
  
  return sets;
}

std::vector<int> getSplit(const std::vector< std::vector<int> >& setToSplit, uint exclude) {
  int count = 0;
  
  for (uint i = 0; i < setToSplit.size(); i++) {
	if (i != exclude) {
	  count += setToSplit[i].size();
	}
  }
  
  std::vector<int> setData(count);
  int a = 0;
  for (uint i = 0; i < setToSplit.size(); i++) {
	if (i != exclude) {
	  for (auto d : setToSplit[i])  {
		setData[a++] = d;
	  }
	}
  }
  
  return setData;
}

double formatError(int correct, size_t testSize) {
  return (testSize - correct) / (double)testSize;
}

void outputResult(int correct, std::chrono::time_point<std::chrono::system_clock> & startTime, size_t size) {
  double error = formatError(correct, size);
  std::cout << "The error rate was \t" << std::setprecision(3)
	<< error << "\tcorrect: \t" << (1-error) << "\tTime taken: \t"
	<< std::chrono::duration_cast<std::chrono::milliseconds> (std::chrono::system_clock::now() - startTime).count() / 1000.0
	<< " s" << std::endl;
}


void test() {
  std::string path = "/Users/bzcschae/workspace/bossVS_classifier/classification/";
  std::string dataset = "beef";
  
  std::string trainFilePath = path;
  const char* trainFileName = trainFilePath.append(dataset).append("/").append(dataset).append("_TRAIN").c_str();
  loadDatset(trainFileName, trainSamples);
  
  int f = 2;
  int windowLength = 100;
  bool norm = false;
  
  const std::set<double> uniqueLabels = uniqueClassLabels(trainSamples);
  
  std::vector< std::vector<int> > wordsTrain (trainSamples.size());
  std::vector<BagOfPattern> bagTrain(trainSamples.size());
  google::dense_hash_map< double, std::shared_ptr<Statistic> > matrix;
  matrix.set_empty_key(DOUBLE_MAX_VALUE);

  
  std::cout << "Labels" << std::endl;
  for (auto labels : uniqueLabels) {
	std::cout << labels << std::endl;
  }
  
  std::cout << std::endl << "Histogram" << std::endl;
  std::shared_ptr<SFA> signature = createWords(windowLength,
							   f,
							   maxS,
							   trainSamples,
							   wordsTrain,
							   NULL,
							   norm);
  
  signature->printHistogram();
  printArray(signature->bestCoefficients, signature->coefficients);
  
  createBagOfPatterns(bagTrain,
					  trainSamples,
					  wordsTrain,
					  f,
					  maxS);
  
  std::map<int, short> counts;
  for (auto bag : bagTrain) {
	for (auto count : bag.wordFrequencies) {
	  counts[count.first] += count.second;
	}
  }
  
  std::cout << std::endl << "Counts" << std::endl;
  for (auto count : counts) {
	std::cout << count.first << ":" << count.second << " ";
	std::cout << std::endl;
  }
  
  std::vector<int> indicesTest(trainSamples.size());
  for (uint i = 0; i < trainSamples.size(); i++) {
	indicesTest[i]=i;
  }
  
  calcTfIdf(bagTrain,
			&matrix,
//			wordInClassFreq,
			trainSamples,
			indicesTest,
			uniqueLabels,
			wordsTrain,
			f,
			maxS);
  
  std::cout << std::endl << "TF IDF" << std::endl;
  for (auto tf : matrix) {
	std::cout << "Class: " << tf.first << std::endl;
	for (auto stat : tf.second->tfIdf) {
	  std::cout << stat.first << ":" << stat.second << " ";
	}
	std::cout << std::endl;
  }
  
  normalizeStatistics(matrix);
  
  std::cout << std::endl << "TF IDF" << std::endl;
  for (auto tf : matrix) {
	std::cout << "Class: " << tf.first << std::endl;
	for (auto stat : tf.second->tfIdf) {
	  std::cout << stat.first << ":" << stat.second << " ";
	}
	std::cout << std::endl;
  }
  
  Predictions p = predict(trainSamples,
						  indicesTest,
						  bagTrain,
						  matrix);
  
  for (double d : p.labels) {
	std::cout << d << " ";
  }
  
  std::cout << std::endl;
}




int main(int argc, char* argv[]) {
  
#ifdef _OPENMP
//  omp_set_num_threads(1);
//  omp_set_nested(1);
#endif

  std::string paths[]{"/Users/bzcschae/workspace/similarity/datasets/classification/", "classification/","../classification/", "/Users/bzcschae/workspace/bossVS_classifier/classification/"};
  
  std::chrono::time_point<std::chrono::system_clock> startTime = std::chrono::system_clock::now();
  startTime = std::chrono::system_clock::now();
  std::vector<std::string> results;
  
  for (uint i = 0; i < sizeof(datasets) / sizeof(std::string); i++) {
	trainSamples.clear();
	testSamples.clear();
	
	std::string path = "";
	for (auto p : paths) {
	  FILE* f;
	  f = fopen(p.c_str(), "r");
	  if (f!=NULL) {
		path = p;
		std::cout << "Using path " << path << "." << std::endl;
		break;
	  }
	}
	
	std::string dataset = datasets[i];
	
	std::string trainFilePath = path;
	const char* trainFileName = trainFilePath.append(dataset).append("/").append(dataset).append("_TRAIN").c_str();
	loadDatset(trainFileName, trainSamples);
	
	std::string testFilePath = path;
	const char* testFileName = testFilePath.append(dataset).append("/").append(dataset).append("_TEST").c_str();
	loadDatset(testFileName, testSamples);
	
	results.push_back(classify());
  }
  
  std::cout << "Total time taken:\t"
	<< std::chrono::duration_cast<std::chrono::milliseconds>
	(std::chrono::system_clock::now() - startTime).count() / 1000.0
	<< " s" << std::endl;

  std::cout << "Total test time taken:\t"
  << testTime
  << " s" << std::endl;
  
  std::cout << "Total train time taken:\t"
  << trainTime
  << " s" << std::endl;

  
  for (auto s : results) {
	std::cout << s << std::endl;
  }
  
  //    MFT::test();
  //    SFA::test();
  //
  //    test();
  
}
